EXPECTED = {'X683': {'extensibility-implied': False,
          'imports': {},
          'object-classes': {},
          'object-sets': {},
          'types': {'IntegerList1': {'type': 'List1',
                                     'chosen-parameters': ['INTEGER']},
                    'List1': {'members': [{'name': 'elem',
                                           'type': 'ElementTypeParam'},
                                          {'name': 'next',
                                           'optional': True,
                                           'type': 'List1',
                                           'parameters': ['ElementTypeParam']}],
                              'type': 'SEQUENCE',
                              'parameters': ['ElementTypeParam']},
                    'OPTIONALLY-SIGNED': {'members': [{'name': 'unsigned-data',
                                                       'tag': {'number': 0},
                                                       'type': 'ToBeSigned'},
                                                      {'name': 'signed-data',
                                                       'tag': {'number': 1},
                                                       'type': 'SIGNED',
                                                       'parameters': ['ToBeSigned']}],
                                          'type': 'CHOICE',
                                          'parameters': ['ToBeSigned']},
                    'OptionallySignedUTF8String': {'type': 'OPTIONALLY-SIGNED',
                                                   'chosen-parameters': ['UTF8String']},
                    'SIGNED': {'members': [{'name': 'authenticated-data',
                                            'type': 'ToBeSigned'},
                                           {'name': 'authenticator',
                                            'type': 'BIT STRING'}],
                               'type': 'SEQUENCE',
                               'parameters': ['ToBeSigned']},
                    'SignedUTF8String': {'type': 'SIGNED',
                                         'chosen-parameters': ['UTF8String']}},
          'values': {}}}
